/***************************************************************************
                          s_partition_function.cpp  -  description
                             -------------------
    begin                : Wed Mar 15 2006
    copyright            : (C) 2006 by Mirela Andronescu
    email                : andrones@cs.ubc.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>

#include "constants.h"
#include "structs.h"
#include "externs.h"
#include "common.h"
#include "simfold.h"
#include "s_energy_matrix.h"
#include "s_hairpin_loop.h"
#include "s_stacked_pair.h"
#include "s_partition_function.h"
#include "s_min_folding.h"
#include "s_sub_folding.h"
#include "params.h"


s_partition_function::s_partition_function (char *cseq)
// The constructor
{
    int i;
    csequence = cseq;    // just refer it from where it is in memory
    seqlen = strlen (cseq);
    sequence = new int[seqlen];
    if (sequence == NULL) giveup ("Cannot allocate memory", "s_partition_function");
    for (i=0; i < seqlen; i++) sequence[i] = nuc_to_int(csequence[i]);
    
    oneoverRT = -10.0/(1.98717*310.15);
    //oneoverRT = 1000.0/(1.98717*310.15);
    
    index = new int [seqlen];
    // an array with indexes, such that we don't work with a 2D array, but with a 1D array of length (n*(n+1))/2
    int total_length = (seqlen *(seqlen+1))/2;
    index[0] = 0;
    for (int i=1; i < seqlen; i++)
        index[i] = index[i-1]+seqlen-i+1;

    u = new double [total_length];    
    if (u == NULL) giveup ("Cannot allocate memory", "s_partition_function");

    up = new double [total_length];    
    if (up == NULL) giveup ("Cannot allocate memory", "s_partition_function");
    
    s1 = new double [total_length];    
    if (s1 == NULL) giveup ("Cannot allocate memory", "s_partition_function");

    upm = new double [total_length];    
    if (upm == NULL) giveup ("Cannot allocate memory", "s_partition_function");
    
    s2 = new double [total_length];    
    if (s2 == NULL) giveup ("Cannot allocate memory", "s_partition_function");
    
    u1 = new double [total_length];    
    if (u1 == NULL) giveup ("Cannot allocate memory", "s_partition_function");
          
    s3 = new double [total_length];    
    if (s3 == NULL) giveup ("Cannot allocate memory", "s_partition_function");          
    
    p = new double[total_length];
    if (p == NULL) giveup ("Cannot allocate memory", "s_partition_function");

    pm = new double[total_length];
    if (pm == NULL) giveup ("Cannot allocate memory", "s_partition_function");
    
    pm1 = new double[total_length];
    if (pm1 == NULL) giveup ("Cannot allocate memory", "s_partition_function");        
        
    // if we want to go the exhaustive way, just for verification
    pexhaustive = new double[total_length];
    if (pexhaustive == NULL) giveup ("Cannot allocate memory", "s_partition_function");        
    
    upexhaustive = new double[total_length];
    if (upexhaustive == NULL) giveup ("Cannot allocate memory", "s_partition_function");        
    
    GlogZ = NULL;
    GlogZexhaustive = NULL;
    
    initialize_arrays ();    
}


s_partition_function::~s_partition_function ()
// The destructor
{
    delete [] sequence;
    delete [] index;     
    delete [] u;
    delete [] up;
    delete [] s1;
    delete [] upm;
    delete [] s2;
    delete [] u1;
    delete [] s3;
    delete [] p;
    delete [] pm;
    delete [] pm1;
    
    // if we go the exhaustive way
    delete [] pexhaustive;
    delete [] upexhaustive;
    
    if (GlogZ != NULL)
        delete [] GlogZ;
        
    if (GlogZexhaustive != NULL)
        delete [] GlogZexhaustive;        
}


void s_partition_function::initialize_arrays ()
{
    int i, j, ij;
    for (i = 0; i < seqlen; i++)
    {
        for (j = i; j < seqlen; j++)
        //for (j = i; j <= MIN(i + TURN, seqlen-1); j++)
        {
            ij = index[i] + j - i;
            u[ij] = 1;
            up[ij] = 0;
            u1[ij] = 0;
            s1[ij] = 0;
            s2[ij] = 0;
            s3[ij] = 0;
            p[ij] = 0;
            pm[ij] = 0;
            pm1[ij] = 0;
            
            // if we go the exhaustive way
            pexhaustive[ij] = 0;
            upexhaustive[ij] = 0;
        }
    }
}


double s_partition_function::compute_partition_function ()
// the recursions are taken from Ding and Lawrence, "A statistical sampling algorithm for RNA secondary structure prediction", NAR 2003
{
    int i, j;
    for (j=TURN+1; j < seqlen; j++)
    {
        for (i=j-TURN-1; i>=0; i--)
        {
            //if (i==40 && j == 53) 
            //{ printf ("Stop\n"); }
            compute_upm(i, j);    // doesn't matter where it is, all dependencies have been computed at previous steps
            compute_u1 (i, j);    // doesn't matter where it is, all dependencies have been computed at previous steps
            compute_up (i, j);    // must be after upm
            compute_u  (i, j);    // must be after up
            compute_s1 (i, j);    // doesn't matter where it is, all dependencies have been computed at previous steps
            compute_s2 (i, j);    // doesn't matter where it is, all dependencies have been computed at previous steps
            compute_s3 (i, j);    // doesn't matter where it is, all dependencies have been computed at previous steps
            //printf ("u(%d,%d) = %Le, up(%d,%d) = %Le\n", i, j, u[index[i]+j-i], i, j, up[index[i]+j-i]);
        }
    }    

    int firstlast = index[0]+seqlen-1;                    
    Z = u[firstlast];
    return u[firstlast];
}


double s_partition_function::exp_AUpenalty (int i, int j)
{
    double AUpen = AU_penalty (sequence[i], sequence[j]);
    return (double)(exp (AUpen * oneoverRT));
}


double s_partition_function::exp_dangle5 (int i, int j, int k)
// dangle_bot
{    
    // added the next lines on Oct 7, 2006
    // make sure we are not out of bounds
    if (i < 0 || i >= seqlen || j < 0 || j >= seqlen || k < 0 || k >= seqlen)
        return 1.0;    
    double dang = dangle_bot[sequence[i]][sequence[j]][sequence[k]];
    return (double) (exp (dang * oneoverRT));
}


double s_partition_function::exp_dangle3 (int i, int j, int k)
// dangle_top
{
    if (i < 0 || i >= seqlen || j < 0 || j >= seqlen || k < 0 || k >= seqlen)
        return 1.0;        
    double dang = dangle_top[sequence[i]][sequence[j]][sequence[k]];
    return (double) (exp (dang * oneoverRT));
}


void s_partition_function::compute_u (int i, int j)
{
    int ij = index[i]+j-i;
    int hj, il, lplus2j, lplus1j;
    int h, l;    
    
    u[ij] = 1 + up[ij] * exp_AUpenalty (i,j);
    
    for (h = i+1; h < j; h++)            
    {
        hj = index[h]+j-h;
        u[ij] += up[hj] * exp_dangle5 (j, h, h-1) * exp_AUpenalty (h, j);
    }                
    for (l = i+1; l < j; l++)            
    {
        il = index[i]+l-i;
        if (l+2 < j)
        {
            lplus2j = index[l+2]+j-l-2;
            lplus1j = index[l+1]+j-l-1;
            u[ij] += up[il] * exp_AUpenalty (i, l) * (exp_dangle3 (l, i, l+1) * u[lplus2j] + u[lplus1j] - u[lplus2j]);
        }
        else
        {
            u[ij] += up[il] * exp_AUpenalty (i, l) * exp_dangle3 (l, i, l+1);
        }            
    }                                   
    for (h = i+1; h < j-1; h++) 
    {
        hj = index[h]+j-h;
        u[ij] += s1[hj];
    }
}



void s_partition_function::compute_up (int i, int j)
{
    int ij = index[i]+j-i;
    int iplus1jminus1 = index[i+1]+j-1-i-1;
    int ip, jp, minq;
    int en_hairpin, en_stack, en_internal;
    
    up[ij] = 0;

    //if (i==0 && j == 8)
    //    printf ("Stop\n");
        
    if (! can_pair (sequence[i], sequence[j]))
    {
        return;
    }    
    
    // hairpin loop
    en_hairpin = s_hairpin_loop::get_energy (i, j, sequence, csequence);
    if (en_hairpin >= INF/2)
        printf ("** Infinite hairpin (%d, %d) !\n", i, j);
    else
        up[ij] += exp (en_hairpin * oneoverRT);                   
        
    // stack pair
    if (can_pair (sequence[i+1], sequence[j-1]))    
    {
        en_stack = s_stacked_pair::get_energy (i, j, sequence);    
        if (en_stack >= INF/2)
            printf ("** Infinite stack   (%d, %d) !\n", i, j);
        else            
            up[ij] += exp (en_stack * oneoverRT) * up [iplus1jminus1];
    }             
             
    // internal loop/bulge
    for (ip = i+1; ip <= MIN(j-2-TURN,i+MAXLOOP+1) ; ip++)  // j-2-TURN
    {
        minq = MAX (j-i+ip-MAXLOOP-2, ip+1+TURN);    // ip+1+TURN);
        for (jp = minq; jp < j; jp++)
        {        
            if (sequence[ip]+sequence[jp] == 3 ||
                sequence[ip]+sequence[jp] == 5)        
            {
                if (ip == i+1 && jp == j-1) continue;    // we don't want stacked pairs here
                int ipjp = index[ip] + jp - ip;
                en_internal = s_internal_loop::get_energy (i, j, ip, jp, sequence);
                if (en_internal >= INF/2)
                    printf ("** Infinite internal(%d, %d) !\n", i, j);
                up[ij] += exp (en_internal * oneoverRT) * up[ipjp]; 
            }
        }        
    }
    
    // multi-loop
    up[ij] += upm[ij];    
}


void s_partition_function::compute_upm (int i, int j)
{
    int ij = index[i]+j-i;
    int l, h;
    int iplus1l, iplus2l, lplus2jminus1, lplus1jminus1, hj;
    double upm_temp;
    upm[ij] = 0;
    for (l=i+2; l < j-3; l++)
    {
        iplus1l = index[i+1]+l-i-1;
        lplus2jminus1 = index[l+2]+j-1-l-2;
        lplus1jminus1 = index[l+1]+j-1-l-1;
        // added / 100.0 on Oct 20
        upm[ij] += up[iplus1l] * exp ((misc.multi_offset + 2*misc.multi_helix_penalty) * oneoverRT) * exp_AUpenalty (i+1,l) *
                   ( exp_dangle3 (l, i+1, l+1) * u1[lplus2jminus1] + u1[lplus1jminus1] - u1[lplus2jminus1]);
    }
    for (l=i+3; l < j-3; l++)
    {
        iplus2l = index[i+2]+l-i-2;
        lplus2jminus1 = index[l+2]+j-1-l-2;
        lplus1jminus1 = index[l+1]+j-1-l-1;  
        // added / 100.0 on Oct 20
        upm[ij] += up[iplus2l] * exp ((misc.multi_offset + 2*misc.multi_helix_penalty + misc.multi_free_base_penalty) * oneoverRT) *
                   exp_dangle3 (i, j, i+1) * exp_AUpenalty (i+2, l) * 
                   (exp_dangle3 (l, i+2, l+1) * u1[lplus2jminus1] + u1[lplus1jminus1] - u1[lplus2jminus1]);
    }
    upm_temp = 0;
    for (h=i+3; h < j-1; h++)
    {
        hj = index[h]+j-h;
        // added / 100.0 on Oct 20
        upm_temp += exp ((misc.multi_offset + 2*misc.multi_helix_penalty + (h-i-1)*misc.multi_free_base_penalty) * oneoverRT) * s2[hj];
    }    
    upm[ij] += exp_dangle3 (i, j, i+1) * upm_temp;
}


void s_partition_function::compute_u1 (int i, int j)
{
    int ij, il, l, h, lplus1j, lplus2j, iplus1l, hj;
    ij = index[i]+j-i;
    u1[ij] = 0;
    for (l=i+1; l <= j; l++)
    {
        il = index[i]+l-i;
        // added / 100.0 on Oct 20
        u1[ij] += up[il] * exp ((misc.multi_helix_penalty) * oneoverRT) * exp_AUpenalty(i,l) *
                  ( f(j+1,i,l) * exp ((j-l) * (misc.multi_free_base_penalty) * oneoverRT) );
                  
        if (l+2 < j)
        {          
            lplus1j = index[l+1]+j-l-1;
            lplus2j = index[l+2]+j-l-2;                  
            // added / 100.0 on Oct 20
            u1[ij] += up[il] * exp ((misc.multi_helix_penalty) * oneoverRT) * exp_AUpenalty(i,l) *
                        (exp_dangle3 (l, i, l+1) * u1[lplus2j] + u1[lplus1j] - u1[lplus2j]);  
        }                        
    }
    for (l=i+2; l <= j; l++)
    {
        iplus1l = index[i+1]+l-i-1;

        // added / 100.0 on Oct 20
        u1[ij] += up[iplus1l] * exp ((misc.multi_helix_penalty + misc.multi_free_base_penalty) * oneoverRT) * exp_AUpenalty(i+1,l) *
                  ( f(j+1,i+1,l) * exp ((j-l) * (misc.multi_free_base_penalty) * oneoverRT)); 

        if (l+2 < j)
        {                                              
            lplus1j = index[l+1]+j-l-1;
            lplus2j = index[l+2]+j-l-2;                  
            // added / 100.0 on Oct 20
            u1[ij] += up[iplus1l] * exp ((misc.multi_helix_penalty + misc.multi_free_base_penalty) * oneoverRT) * exp_AUpenalty(i+1,l) *       
                    ( exp_dangle3 (l, i+1, l+1) * u1[lplus2j] + u1[lplus1j] - u1[lplus2j]);      
        }                    
    }
    for (h=i+2; h <= j-1; h++)
    {
        hj = index[h]+j-h;
        // added / 100.0 on Oct 20
        u1[ij] += exp ((misc.multi_helix_penalty + (h-i)*misc.multi_free_base_penalty) * oneoverRT) * s3[hj];
    }
}


void s_partition_function::compute_s1 (int h, int j)
{
    int hj, hl, l, lplus2j, lplus1j;
    hj = index[h]+j-h;
    s1[hj] = 0;
    
    for (l = h+1; l < j; l++)
    {
        hl = index[h] + l - h;
        if (l+2 < j)
        {
            lplus2j = index[l+2]+j-l-2;
            lplus1j = index[l+1]+j-l-1;   
            
            s1[hj] += up[hl] * exp_dangle5 (l, h, h-1) * exp_AUpenalty (h, l) * 
                    ( exp_dangle3 (l, h, l+1) * u[lplus2j] + u[lplus1j] - u[lplus2j] );
        }
        else
        {
            s1[hj] += up[hl] * exp_dangle5 (l, h, h-1) * exp_AUpenalty (h, l) * exp_dangle3 (l, h, l+1);
        
        }                    
    }
}


void s_partition_function::compute_s2 (int h, int j)
{
    int hj, hl, l, lplus2jminus1, lplus1jminus1;
    hj = index[h]+j-h;
    s2[hj] = 0;
    
    for (l = h+1; l < j-3; l++)
    {
        hl = index[h] + l - h;
        lplus2jminus1 = index[l+2]+j-1-l-2;
        lplus1jminus1 = index[l+1]+j-1-l-1;        
        s2[hj] += up[hl] * exp_dangle5 (l, h, h-1) * exp_AUpenalty (h, l) * 
                  ( exp_dangle3 (l, h, l+1) * u[lplus2jminus1] + u[lplus1jminus1] - u[lplus2jminus1] );
    }
}

double s_partition_function::f (int jplus1, int h, int l)
{
    if (l > jplus1-1)    giveup ("Error, l > j", "f function, partition_function");
    if (l == jplus1-1)
        return 1.0;
    return exp_dangle3 (l, h, l+1);    
}



void s_partition_function::compute_s3 (int h, int j)
{
    int hj, hl, l, lplus2j, lplus1j;
    hj = index[h]+j-h;
    s3[hj] = 0;
    
    for (l = h+1; l <= j; l++)
    {
        hl = index[h] + l - h;
        // added / 100.0 on Oct 20
        s3[hj] += up[hl] * exp_dangle5 (l, h, h-1) * exp_AUpenalty (h, l) * 
                  ( f (j+1, h, l) * exp ((j-l) * misc.multi_free_base_penalty * oneoverRT) );
        
        if (l+2 < j)                  
        {
            lplus2j = index[l+2]+j-l-2;
            lplus1j = index[l+1]+j-l-1;                          
            s3[hj] += up[hl] * exp_dangle5 (l, h, h-1) * exp_AUpenalty (h, l) *                   
                    ( exp_dangle3 (l, h, l+1) * u1[lplus2j] + u1[lplus1j] - u1[lplus2j] );
        }                    
    }
}


void s_partition_function::compute_base_pair_probabilities ()
// Nov 9, 2006. Computes base pair probabilities
// PRE: the partition function arrays have been filled
{
    int h, l;
    for (h=0; h < seqlen; h++)   
    {
        for (l=seqlen-1; l > h+TURN; l--)
        {
            if (can_pair (sequence[h], sequence[l]))    
            {
                compute_p (h, l);
                compute_pm (h, l);
                compute_pm1 (h, l);            
            }
        }
    }
}


void s_partition_function::compute_p (int h, int l)
{
    int hl, zeronminus1, hminus1lplus1;
    int en_stack;
    int i,j;
    int en_internal;
    double term1, term2;
    
    hl = index[h]+l-h;

    //exterior base pair
    // first, 5' end
    term1 = 0;
    if (h > TURN)
    {
        term1 += u[index[0]+h-1] - u[index[0]+h-3] + exp_dangle5(l,h,h-1) * u[index[0]+h-3];
        //printf ("Add1 d5(%d,%d,%d), term1=%g\n", l, h, h-1, term1);
    }
    else if (h > 0)       
    {
        term1 += exp_dangle5(l,h,h-1);  
        //printf ("Add2 d5(%d,%d,%d)\n", l, h, h-1);
    }
    else                  term1 = 1;  
    
    // then, 3' end
    term2 = 0;
    if (l < seqlen-3)     
    {
        term2 += u[index[l+1]+seqlen-1-(l+1)] - u[index[l+2]+seqlen-1-(l+2)] + exp_dangle3(l,h,l+1) * u[index[l+2]+seqlen-1-(l+2)];    // lplus1nminus1 (l+1,n-1)
        //printf ("Add3 d3(%d,%d,%d), term2=%g\n", l, h, l+1, term2);
    }
    else if (l < seqlen-1) 
    {    
        term2 += exp_dangle3(l,h,l+1);
        //printf ("Add4 d3(%d,%d,%d)\n", l, h, l+1);
    }
    else                  term2 = 1;
    
    zeronminus1 = index[0] + seqlen-1;    
    p[hl] = term1 * up[hl] * term2 / u[zeronminus1] * exp_AUpenalty (h, l);
    //if (h > 0)             p[hl] *= exp_dangle5 (l, h, h-1);
    //if (l < seqlen - 1)    p[hl] *= exp_dangle3 (l, h, l+1);
    
    // case stacking energies
    if (h > 0 && l < seqlen-1)
    {
        hminus1lplus1 = index[h-1] + l+1 - h+1;
        if (can_pair (sequence[h-1], sequence[l+1]))    
        {
            en_stack = s_stacked_pair::get_energy (h-1, l+1, sequence);    
            if (en_stack >= INF/2)
                printf ("** Infinite stack   (%d, %d) !\n", h-1, l+1);
            else            
                p[hl] += p[hminus1lplus1] * up[hl] / up[hminus1lplus1] * exp (en_stack * oneoverRT);
        }
    }
    
    // case internal loop
    for (i = MAX(0, h-MAXLOOP-1); i < h; i++)
    {
        for (j = l+1; j <= MIN(MAXLOOP+l-h+i-2, seqlen-1); j++)
        {        
            if (sequence[i]+sequence[j] == 3 ||
                sequence[i]+sequence[j] == 5)        
            {
                if (i == h-1 && j == l+1) continue;    // we don't want stacked pairs here
                int ij = index[i] + j - i;
                en_internal = s_internal_loop::get_energy (i, j, h, l, sequence);
                if (en_internal >= INF/2)
                    printf ("** Infinite internal(%d, %d) !\n", i, j);
                p[hl] += p[ij] * up[hl] /up[ij] * exp (en_internal * oneoverRT); 
            }
        }        
    }
    
    // case multi-loop
    for (i=0; i < h; i++)
    {
        int il = index[i] + l - i;
        int iplus1hminus1 = index[i+1] + h-1 - (i+1);
        p[hl] += up[hl] * exp ((misc.multi_offset + 2*misc.multi_helix_penalty) * oneoverRT) * (pm1[il]*u1[iplus1hminus1] + pm[il] * (u1[iplus1hminus1] + exp (((h-i-1)*misc.multi_free_base_penalty) * oneoverRT)));
    }    
}
/*

void s_partition_function::compute_pm_nod (int i, int l)
// no d3 has to be added to the right of l
{
    int j;
    int ij, lplus1jminus1, il;
    il = index[i]+l-i;   
    for (j = l+4; j < seqlen-1; j++)
    {
        ij = index[i]+j-i;   
        lplus1jminus1 = index[l+1] + j-1 - l-1;
        if (up[ij] != 0)
            pm[il] += p[ij] * (u1[lplus1jminus1] - u1[lplus2jminus1]) / up[ij];
    }
}


void s_partition_function::compute_pm_d3 (int i, int l)
// d3 has to be added to the right of l
{
    int j;
    int ij, lplus1jminus1, il;
    il = index[i]+l-i;   
    for (j = l+4; j < seqlen-1; j++)
    {
        ij = index[i]+j-i;   
        lplus1jminus1 = index[l+1] + j-1 - l-1;
        if (up[ij] != 0)
            pm[il] += p[ij] * u1[lplus2jminus1] / up[ij];
    }
}


*/

void s_partition_function::compute_pm (int i, int l)
{
    int j;
    int ij, lplus1jminus1, il;
    il = index[i]+l-i;   
    for (j = l+1; j < seqlen; j++)
    {
        ij = index[i]+j-i;   
        lplus1jminus1 = index[l+1] + j-1 - l-1;
        if (up[ij] != 0)
            pm[il] += p[ij] * u1[lplus1jminus1] / up[ij];
    }
}

void s_partition_function::compute_pm1 (int i, int l)
// region l+1 ... j-1 is unpaired
{
    int j;
    int ij, il;
    il = index[i]+l-i;   
    for (j = l+1; j < seqlen; j++)
    {
        ij = index[i]+j-i;   
        if (up[ij] != 0)
            pm1[il] += p[ij] / up[ij] * exp ((j-l-1) * misc.multi_free_base_penalty * oneoverRT);
    }
}


/*---------------------------------------------------------------------------*/
#define PMIN 0.00001
void s_partition_function::PS_dot_plot(char *wastlfile)
// taken from Vienna package 1.4 
// plots a matrix with the square root of probabilities (for better visibility)
{
  /* produce PostScript dot plot from probabilities in p[] array */
   
  FILE *wastl;
  char name[31], *c;
  int i, j, length;
  double tmp;
   
  length= strlen(csequence);
  wastl = fopen(wastlfile,"w");
  if (wastl==NULL) {
    fprintf(stderr, "can't open %s for dot plot\n", wastlfile);
    return; /* return 0 for failure */
  }
  strncpy(name, wastlfile, 30);
  if ((c=strrchr(name, '_'))!=0) *c='\0';
  fprintf(wastl,"%%!PS-Adobe-3.0 EPSF-3.0\n");
  fprintf(wastl,"%%%%Title: RNA DotPlot\n");
  fprintf(wastl,"%%%%Creator: PS_dot.c, ViennaRNA Package\n");
  //fprintf(wastl,"%%%%CreationDate: %s", time_stamp());
  fprintf(wastl,"%%%%BoundingBox: 66 211 518 662\n");
  fprintf(wastl,"%%%%DocumentFonts: Helvetica\n");
  fprintf(wastl,"%%%%Pages: 1\n");
  fprintf(wastl,"%%%%EndComments\n\n");
//  fprintf(wastl,"%%Options: %s\n", option_string());

  fprintf(wastl,"%%This file contains the square roots "
      "of the base pair probabilities in the form\n");
  fprintf(wastl,"%% i  j  sqrt(p(i,j)) ubox\n");
   
  fprintf(wastl,"100 dict begin\n");  /* DSC says EPS should create a dict */
  fprintf(wastl,"\n/logscale false def\n\n");
  fprintf(wastl,"%%delete next line to get rid of title\n"
      "270 665 moveto /Helvetica findfont 14 scalefont setfont "
      "(%s) show\n\n", name);
  fprintf(wastl,"/lpmin {\n"
      "   %g log  %% log(pmin) only probs>pmin will be shown\n"
      "} bind def\n\n",PMIN);

  /* EPS should not contain lines >255 characters */
  fprintf(wastl,"/sequence { (\\\n");
  i=0;
  while (i<length) {
    fprintf(wastl, "%.255s\\\n", csequence+i);
    i+=255;
  }
  fprintf(wastl,") } def\n");
  fprintf(wastl,"/len { sequence length } def\n\n");
   
  fprintf(wastl,"/ubox {\n"     /* upper triangle matrix */
      "   logscale {\n"
      "      log dup add lpmin div 1 exch sub dup 0 lt { pop 0 } if\n"
      "   } if\n"
      "   3 1 roll\n"
      "   exch len exch sub 1 add box\n"
      "} bind def\n\n");

  fprintf(wastl,"/lbox {\n"     /* lower triangle matrix */
      "   3 1 roll\n"
      "   len exch sub 1 add box\n"
      "} bind def\n\n");

  fprintf(wastl,"/box { %%size x y box - draws box centered on x,y\n"
      "   2 index 0.5 mul add            %% x += 0.5\n"
      "   exch 2 index 0.5 mul add exch  %% x += 0.5\n"
      "   newpath\n"
      "   moveto\n"
      "   dup neg   0 rlineto\n"
      "   dup neg   0 exch rlineto\n"
      "             0 rlineto\n"
      "   closepath\n"
      "   fill\n"
      "} def\n\n");

  fprintf(wastl,"72 216 translate\n");
  fprintf(wastl,"72 6 mul len 1 add div dup scale\n");
  fprintf(wastl,"/Helvetica findfont 0.95 scalefont setfont\n\n");

   /* print sequence along all 4 sides */
  fprintf(wastl,"%% print sequence along all 4 sides\n");
  fprintf(wastl,"0 1 len 1 sub {\n");
  fprintf(wastl,"    dup\n");
  fprintf(wastl,"    0.7 add -0.3 moveto\n");
  fprintf(wastl,"    sequence exch 1 getinterval\n");
  fprintf(wastl,"    show\n");
  fprintf(wastl,"} for\n");
  fprintf(wastl,"\n");
  fprintf(wastl,"0 1 len 1 sub {\n");
  fprintf(wastl,"    dup\n");
  fprintf(wastl,"    0.7 add 0.7 len add moveto\n");
  fprintf(wastl,"    sequence exch 1 getinterval\n");
  fprintf(wastl,"    show\n");
  fprintf(wastl,"} for\n\n");

  fprintf(wastl,"90  rotate\n");
  fprintf(wastl,"0 1 len 1 sub {\n");
  fprintf(wastl,"    dup\n");
  fprintf(wastl,"    0.7 add -0.2 moveto\n");
  fprintf(wastl,"    len 1 sub exch sub\n");
  fprintf(wastl,"    sequence exch 1 getinterval\n");
  fprintf(wastl,"    show\n");
  fprintf(wastl,"} for\n");
  fprintf(wastl,"270 rotate\n\n");

  fprintf(wastl,"270 rotate\n");
  fprintf(wastl,"0 1 len 1 sub {\n");
  fprintf(wastl,"    dup\n");
  fprintf(wastl,"    -0.3 add len sub  0.7 len add  moveto\n");
  fprintf(wastl,"    sequence exch 1 getinterval\n");
  fprintf(wastl,"    show\n");
  fprintf(wastl,"} for\n");
  fprintf(wastl,"90 rotate\n\n");

  /* do grid */
  fprintf(wastl,"0.5 dup translate\n"
      "%% draw diagonal\n"
      "0.04 setlinewidth\n"
      "0 len moveto len 0 lineto stroke \n\n");
  fprintf(wastl,"%%draw grid\n"
      "0.01 setlinewidth\n"
      "len log 0.9 sub cvi 10 exch exp  %% grid spacing\n"
      "dup 1 gt {\n"
      "   dup dup 20 div dup 2 array astore exch 40 div setdash\n"
      "} { [0.3 0.7] 0.1 setdash } ifelse\n"
      "0 exch len {\n"      /* for (i=0; i<=len; i++) */
      "   dup dup\n"        
      "   0 moveto\n"                     /* i 0 moveto   */
      "   len lineto \n"                  /* i len lineto */
      "   dup\n"
      "   len exch sub 0 exch moveto\n"   /* 0 i moveto   */
      "   len exch len exch sub lineto\n" /* len i lineto */
      "   stroke\n"
      "} for\n"
      "0.5 neg dup translate\n\n");

  /* print boxes */
  for (i=0; i<length; i++)
    for (j=i+1; j<=length; j++) {
      if (p[index[i]+j-i]<PMIN) continue;
      tmp = sqrt(p[index[i]+j-i]);
      fprintf(wastl,"%d %d %1.5f ubox\n", i+1, j+1, tmp);
    }
    // come back to this later    
  /* do mfe */
  /*
  if (base_pair)
    for(i=1; i<=base_pair[0].i; i++) 
      fprintf(wastl,"%d %d 0.95 lbox\n",
          base_pair[i].i, base_pair[i].j); 
  */  
  fprintf(wastl,"showpage\n");
  fprintf(wastl,"end\n");
  fprintf(wastl,"%%%%EOF\n");
  fclose(wastl);
  //return 1; /* success */
}


int s_partition_function::has_base_pair (int i, int j, char *structure)
// returns 1 if structure contains base pair (i,j)
{
    int ptable[MAXSLEN];
    detect_original_pairs (structure, ptable); 
    if (ptable[i] == j)
    {
        //printf ("Structure %s has base pair %d-%d\n", structure, i, j);    
        return 1;
    }
    return 0;
}


int s_partition_function::identical_structure (int i, int j, char *structure1, char *structure2)
// return 1 if structure 1 and structure2 are identical between i and j
{
    int k;
    for (k=i; k <= j; k++)
    {
        if (structure1[k] != structure2[k])
            return 0;
    }
    return 1;
}

double s_partition_function::exp_free_energy_partial (int i, int j, char *sequence, char *structure)
// return exp of free energy of sequence and structure, for region i-j, if i and j are paired with each other
// if i and j are not paired with each other, return 0
{
    char partial_sequence[MAXSLEN];
    char partial_structure[MAXSLEN];
    int k;
    double energy, beta;
    beta = 1000.0/(1.98717*310.15);
    if (has_base_pair (i,j, structure))
    {        
        for (k=i; k <= j; k++)
        {
            partial_sequence[k-i] = sequence[k];
            partial_structure[k-i] = structure[k];
        }
        partial_sequence[k-i] = '\0';
        partial_structure[k-i] = '\0';
        //printf ("Partial: %s %s\n", partial_sequence, partial_structure);
        energy = free_energy_simfold (partial_sequence, partial_structure);
        return exp(-1.0 * energy * beta);
    }
    return 0.0;
}



double s_partition_function::compute_partition_function_exhaustively ()
{
    char structure[MAXSLEN];
    double enthalpy, energy;
    char tmp_structures[MAXSUBSTR][MAXSLEN];
    double tmp_energies[MAXSUBSTR];
    double min_energy, max_energy;
    int actual_num_str;
    int i,ii,j, k;
    double R, temp, beta;
    //R = 0.00198717;
    //temp = 310.15;
    beta = 1000.0/(1.98717*310.15);
    //beta = 1; 

    simple_dangling_ends = 1;
    int ptable[MAXSLEN];
    double strprob;
        
    s_min_folding *min_fold = new s_min_folding (csequence);
    min_energy = min_fold->s_simfold ();
    min_fold->return_structure (structure);
    delete min_fold;      
    
    //s_sub_folding* sub_fold = new s_sub_folding(sequence, -(int)(min_energy*100.0));
    s_sub_folding* sub_fold = new s_sub_folding(csequence, 10000);
    sub_fold->set_limit (MAXSUBSTR);
    sub_fold->s_simfold (enthalpy);
    actual_num_str = sub_fold->return_structures(tmp_structures, tmp_energies);
    delete sub_fold;
    
    max_energy = tmp_energies[actual_num_str-1];
    Zexhaustive = 0.0;
    for (i=0; i < actual_num_str; i++)
    {
        // recompute the free energy, i.e. with the correct dangling ends
        energy = free_energy_simfold (csequence, tmp_structures[i]);
        strprob = (double)(exp(-1.0 * energy * beta)/Z);
        detect_original_pairs (tmp_structures[i], ptable); 
        for (ii=0; ii < seqlen; ii++)
        {
            if (ptable[ii] > ii)
            {
                //printf ("ptable[%d]=%d\n", ii, ptable[ii]);
                int ind = index[ii] + ptable[ii] - ii;
                pexhaustive[ind] += strprob;
            }
        }   
        
        printf ("Substr %d: %s\ten=%.2lf \tprob=%e\n", i, tmp_structures[i], energy, strprob);
        Zexhaustive += exp (-1.0 * energy * beta);
    }
    
    // try to get the exhaustive up[i,j]
    int l;
    for (i=0; i < seqlen; i++)
    {
        for (j=i+TURN+1; j < seqlen; j++)
        {    
            int ij = index[i] + j - i;
            if (can_pair(sequence[i], sequence[j]))
            {
                // compute exhaustive up[i,j]
                for (k=0; k < actual_num_str; k++)                
                {
                    if (has_base_pair(i,j, tmp_structures[k]))
                    {
                        int identical = 0;
                        for (l=0; l < k; l++)
                        {
                            if (identical_structure (i, j, tmp_structures[l], tmp_structures[k]))
                                identical = 1;
                        }
                        if (!identical)
                            upexhaustive[ij] += exp_free_energy_partial (i, j, csequence, tmp_structures[k]);
                        //else
                        //    printf ("Str %s identical with other for i=%d, j=%d\n", tmp_structures[k], i, j);
                    }                        
                }
            }
        }
    }
    return Zexhaustive;        
}


void s_partition_function::verify_partition_function()
{
    //#define PREC 0.0000000001
    #define PREC 1.0e-15
    int i,j,ij;    
    printf ("Checking partition function... ");
    if (Z - Zexhaustive < PREC && Zexhaustive - Z < PREC)
        printf ("\t++++++++++++++++++++++\n");
    else
        printf ("\t----------------------\n\tZ = %g, Ze = %g, diff = %g\n", Z, Zexhaustive, Z-Zexhaustive);
        
        
    printf ("Checking up .................. ");
    int upfine = 1;
    for (i=0; i < seqlen; i++)
        for (j=i+1; j < seqlen; j++)
        {
            ij = index[i]+j-i;
            if (up[ij] - upexhaustive[ij] > PREC || upexhaustive[ij] - up[ij] > PREC)
            {
                if (upfine)
                    printf ("\t----------------------\n");
                printf ("\tup[%d,%d] = %g, upex = %g, diff = %g\n", i,j,up[ij],upexhaustive[ij], up[ij]-upexhaustive[ij]);
                upfine = 0;
            }
        }
    if (upfine) 
        printf ("\t++++++++++++++++++++++\n");
        
        
    printf ("Checking p ................... ");
    int pfine = 1;
    for (i=0; i < seqlen; i++)
        for (j=i+1; j < seqlen; j++)
        {
            ij = index[i]+j-i;
            if (p[ij] - pexhaustive[ij] > PREC || pexhaustive[ij] - p[ij] > PREC)
            {
                if (pfine)
                    printf ("\t----------------------\n");
                printf ("\tp[%d,%d] = %g, pex = %g, diff = %g\n", i,j,p[ij],pexhaustive[ij], p[ij]-pexhaustive[ij]);
                pfine = 0;
            }
        }
    if (pfine) 
        printf ("\t++++++++++++++++++++++\n");

}



void s_partition_function::compute_logZ_gradient_exhaustively ()
{
    num_params = create_string_params();
    if (GlogZexhaustive == NULL)
    {        
        GlogZexhaustive = new double[num_params];
    }    
    
    char structure[MAXSLEN];
    double enthalpy, energy;
    char tmp_structures[MAXSUBSTR][MAXSLEN];
    double tmp_energies[MAXSUBSTR];
    double min_energy, max_energy;
    int i, k;
    double R, temp, beta;
    int real_str_found, actual_num_str;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp);
    real_str_found = 0;
    double numerator [MAXNUMPARAMS];
    int counter [MAXNUMPARAMS];
    double denominator;      
    
    s_sub_folding* sub_fold = new s_sub_folding(csequence, 10000);
    sub_fold->set_limit (MAXSUBSTR);
    sub_fold->s_simfold (enthalpy);
    actual_num_str = sub_fold->return_structures(tmp_structures, tmp_energies);
    delete sub_fold;
    
    for (i=0; i < num_params; i++)
    {
        numerator[i] = 0;
    }    
        
    // first compute the denominator, which is the same for all parameters
    denominator = 0;
    for (k=0; k < actual_num_str; k++)
    {
        // recompute the free energy, i.e. with the correct dangling ends
        energy = free_energy_simfold (csequence, tmp_structures[k]);        
        denominator += exp ((-1) * energy * beta);
        count_each_structure_type (csequence, tmp_structures[k], counter, 1);        
        for (i=0; i < num_params; i++)
        {            
            numerator[i] += counter[i] * exp ((-1) * energy * beta);
        }
        //printf ("Sub str %d, counter[i] = %d, numerator[i] = %e\n", k, counter[7646], numerator[7646]);
    }
    // now the denominator and nominator are computed
           
    for (i=0; i < num_params; i++)
    {
        GlogZexhaustive[i] = numerator[i] / denominator;
    }      
    //printf ("logZ_gradient[i] = %e\n", logZ_gradient[7646]);
    
}


void s_partition_function::compute_logZ_gradient ()
// u, up and p arrays are filled
{ 

    int index_param;
    int i, j, k, l;
    int ii, jj, iip, jjp, iijj, iipjjp, iiplus1jjminus1;
    int en_hairpin, en_stack, en_internal;
    
    num_params = create_string_params();
    if (GlogZ == NULL)
    {        
        GlogZ = new double[num_params];        
    }       
    
    for (i=0; i < num_params; i++)    GlogZ[i] = 0;
    
    // stacking energies
    index_param = 0;
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                {
                    if (stack[i][j][k][l] < INF)
                    {
                        // exclude duplicates
                        // stack[i][j][k][l] is the same as stack[l][k][j][i]
                        // TODO: to come back to duplicates
                        if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                        {
                            // traverse the whole sequence and look for this building block
                            for (ii = 0; ii < seqlen; ii++)
                            {
                                for (jj = ii+TURN+3; jj < seqlen; jj++)
                                {
                                    // TODO: stack is also used for bulge
                                    if (sequence[ii] == i && sequence[jj] == j && 
                                        sequence[ii+1] == k && sequence[jj-1] == l)
                                    {
                                        iijj = index[ii] + jj -ii;
                                        iiplus1jjminus1 = index[ii+1] + jj-1 - (ii+1);
                                        en_stack = s_stacked_pair::get_energy (ii, jj, sequence);    
                                        GlogZ[index_param] += p[iijj] * up[iiplus1jjminus1] / up[iijj] * exp (en_stack * oneoverRT);                                        
                                    }
                                }                                                
                            }
                            index_param++;                                
                        }                        
                    }                    
                }
    // tstackh energies
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                {
                    if (tstackh[i][j][k][l] < INF)
                    {
                        // no duplicates here
                        for (ii = 0; ii < seqlen; ii++)
                        {
                            // tstackh is not added to hairpin loops of size 3
                            for (jj = ii+TURN+2; jj < seqlen; jj++)
                            {
                                if (sequence[ii] == i && sequence[jj] == j && 
                                    sequence[ii+1] == k && sequence[jj-1] == l)
                                {
                                    iijj = index[ii] + jj -ii;
                                    en_hairpin = s_hairpin_loop::get_energy (ii, jj, sequence, csequence);
                                    GlogZ[index_param] += p[iijj] / up[iijj] * exp (en_hairpin * oneoverRT);   
                                }
                            }
                        }
                        index_param++;
                    }
                }
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                {
                    if (tstacki[i][j][k][l] < INF)
                    {
                        // no duplicates here
                        for (ii = 0; ii < seqlen; ii++)
                        {
                            for (jj = ii+TURN+4; jj < seqlen; jj++)
                            {
                                if (sequence[ii] == i && sequence[jj] == j && 
                                    sequence[ii+1] == k && sequence[jj-1] == l)
                                {
                                    iijj = index[ii] + jj -ii;
                                    
                                    for (iip = ii+1; iip <= MIN(jj-2-TURN,ii+MAXLOOP+1) ; iip++)  // j-2-TURN
                                    {
                                        int minq = MAX (jj-ii+iip-MAXLOOP-2, iip+1+TURN);    // ip+1+TURN);
                                        for (jjp = minq; jjp < jj; jjp++)
                                        {        
                                            if (sequence[iip]+sequence[jjp] == 3 ||
                                                sequence[iip]+sequence[jjp] == 5)        
                                            {
                                            
                                                int branch1 = iip-ii-1;
                                                int branch2 = jj-jjp-1;
                                                if (branch1 < 0 || branch2 < 0)    // not valid
                                                    continue;
                                                if (branch1 == 0 && branch2 == 0)    // stack pair, not good here
                                                    continue;
                                                if (branch1 == 1 && branch2 == 2 && !simple_internal_energy)
                                                    continue;
                                                if (branch1 == 2 && branch2 == 1 && !simple_internal_energy)
                                                    continue;
                                                if (branch1 == 2 && branch2 == 2 && !simple_internal_energy)
                                                    continue;
                                                if (branch1 == 0 || branch2 == 0)    // bulge
                                                    continue;
                                                if ((branch1 == 1 || branch2 == 1) && misc.gail_rule)

                                                iijj = index[ii] + jj -ii;
                                                iipjjp = index[iip] + jjp - iip;                      
                                
                                                if (branch1 != 0 || branch2 != 0)
                                                {
                                                    if (k==0 && l==0)    
                                                    {
                                                        en_internal = s_internal_loop::get_energy (ii, jj, iip, jjp, sequence);
                                                    }
                                                }
                                                else
                                                {
                                                    en_internal = s_internal_loop::get_energy (ii, jj, iip, jjp, sequence);
                                                    GlogZ[index_param] += p[iijj] * up[iipjjp] / up[iijj] * exp (en_internal * oneoverRT);    
                                                }    
                                            }
                                        }
                                    }                                    
                                }
                            }
                        }
                        index_param++;                    
                    }
                }                
                
                
}
                
                
void s_partition_function::verify_gradient ()
{
    int index_param;
    int i, j, k, l, m, n, o, p;
    index_param = 0;
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                {
                    if (stack[i][j][k][l] < INF)
                    {
                        // exclude duplicates
                        // stack[i][j][k][l] is the same as stack[l][k][j][i]
                        // TODO: to come back to duplicates
                        if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                        {
                            printf ("%d stack[%d][%d][%d][%d]: exhaust = %g\tsmart = %g\n", index_param, i, j, k, l, GlogZexhaustive[index_param], GlogZ[index_param]);    
                            index_param++;
                        }
                    }
                }
    // tstackh energies
    for (i=0; i < NUCL; i++)
        for (j=0; j < NUCL; j++)
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                {
                    if (tstackh[i][j][k][l] < INF)
                    {
                        printf ("%d tstackh[%d][%d][%d][%d]: exhaust = %g\tsmart = %g\n", index_param, i, j, k, l, GlogZexhaustive[index_param], GlogZ[index_param]);    
                                    index_param++;
                    }
                }

    /*                            
    for (i=0; i < num_params; i++)
    {
        printf ("GlogZexhaustive[%d] = %g\n", i, GlogZexhaustive[i]);
    }
    */
}



